import numpy as np

def n_kron(gate, n):
    result = gate
    for i in range(n-1):
        result = np.kron(result, gate)
    return result

def gate_builder(gates):
    if not gates:
        raise ValueError('Gate builder expects one or more gates')
    
    result = gates[0]
    for i in range(1, len(gates)):
        result = np.kron(result, gates[i])
    
    return result


import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import numpy as np
import qupyt as qp

def grover(f, num_sols=1, buffer=10):

    def oracle():
        nonlocal buffer
        size = 2 ** buffer
        oracle = np.identity(size)
        for i in range(size):
            oracle[i, i] = -1 if f(i) == 1 else 1
        return oracle

    def diffusion():
        nonlocal buffer
        size = 2 ** buffer
        identity = np.identity(size)
        all_ones = np.ones((size, size))
        diffusion = 2 * all_ones / size - identity
        return diffusion
    
    
    intermediate = []

    state = qp.layers.Input(buffer) # Initializes circuit with n qubits (all set to ket0)


    state = qp.layers.N_Hadamard(state) # Apply Hadamard on all the qubits
    intermediate.append(state)
    

    for i in range(int(np.floor(np.pi/4 * np.sqrt(np.power(2, buffer)/num_sols)))): # Repeat until probability of measuring intended answer is optimized
        state = qp.layers.Custom(state, oracle()) # Apply the oracle to entire circuit
        intermediate.append(state)

        state = qp.layers.Custom(state, diffusion()) # Apply the diffusion to entire circuit
        intermediate.append(state)

    measured = qp.layers.Measure(state, list(range(buffer))) # Measure all qubits to get the answer

    print(f"Grover's Found:      {int(measured[0], 2)} with {state[np.where(measured[1] == 1)[0]][0][0]*100:.02f}% certainty")

    return measured[0], intermediate





def animate_grover(intermediate, BUFFER, MARKED_INDICES):

    def grover_basis(N, marked_indices):
        # Create |alpha⟩ (non-solution states)
        alpha = np.ones(N, dtype=complex)
        for idx in marked_indices:
            alpha[idx] = 0
        alpha /= np.linalg.norm(alpha)
        
        # Create |beta⟩ (solution state(s))
        beta = np.zeros(N, dtype=complex)
        for idx in marked_indices:
            beta[idx] = 1
        beta /= np.linalg.norm(beta)

        return alpha, beta

    def project_to_grover_plane(state, alpha, beta):
        proj_alpha = np.vdot(alpha, state)
        proj_beta = np.vdot(beta, state)
        return np.array([proj_alpha, proj_beta])
    
    alpha, beta = grover_basis(np.power(2, BUFFER), MARKED_INDICES)

    states_projected = np.array([project_to_grover_plane(i, alpha, beta) for i in intermediate])

    fig, ax = plt.subplots(figsize=(12, 12))
    ax.set_xlim(-1, 1)
    ax.set_ylim(-1, 1)
    ax.set_xlabel('Projection on |α⟩ (non-solution space)')
    ax.set_ylabel('Projection on |β⟩ (solution space)')
    ax.axhline(0, color='black', linewidth=0.5)
    ax.axvline(0, color='black', linewidth=0.5)
    ax.grid(True)
    ax.set_title("Grover's Algorithm State Evolution")

    # Initialize the line objects for the animation
    line, = ax.plot([], [], 'bo-', label='State Vector', markersize=8)
    text = ax.text(0, 0, "", color="black", fontsize=12)

    # Initialize the plot
    def init():
        line.set_data([], [])
        text.set_text("")
        return line, text

    # Update function for animation
    def update(frame):
        x, y = states_projected[frame]
        line.set_data(states_projected[:frame+1, 0], states_projected[:frame+1, 1])
        text.set_text(f'Step {frame}')
        return line, text

    # Create animation
    ani = FuncAnimation(fig, update, frames=len(states_projected), init_func=init, blit=True, interval=100)

    #ani.save('grover.gif', writer='pillow', fps=5)
    
    plt.legend()
    plt.show()


import qupyt as qp
import numpy as np

def Input(num_bits):
    return qp.subroutines.n_kron(qp.states.ZERO, num_bits)

def N_Hadamard(state):
    return np.round(qp.subroutines.n_kron(qp.gates.H, int(np.log2(state.shape[0]))) @ state, 10)

def Custom(state, gate):
    return np.round(gate @ state, 10)

def Measure(state, bits):
    return qp.Qubit.partial_measure(state, bits)


import numpy as np
import qupyt as qp

def bernstein_vazirani(f, buffer):

    def bernstein_vazirani_Uf():
        nonlocal f
        nonlocal buffer
        dim = 2 ** (buffer + 1)
        U = np.zeros((dim, dim))

        for i in range(dim):
            bits = format(i, f'0{buffer+1}b')
            x_bits = bits[:buffer]
            y_bit = bits[buffer]
            
            x_int = int(x_bits, 2)
            y_int = int(y_bit)
            
            output_y = y_int ^ f(x_int, buffer)
            
            out_bits = x_bits + str(output_y)
            input_idx = i
            output_idx = int(out_bits, 2)
            
            U[output_idx][input_idx] = 1

        return U

    
    state = qp.layers.Input(buffer+1) # Initializes circuit with n+1 qubits (all set to ket0)
    state = qp.layers.Custom(state, qp.subroutines.gate_builder([qp.subroutines.n_kron(qp.gates.I, buffer), qp.gates.X])) # Set the output qubit to ket1
    state = qp.layers.N_Hadamard(state) # Apply Hadamard on all the qubits
    state = qp.layers.Custom(state, bernstein_vazirani_Uf()) # Apply the Deutsch-Josza unitary to entire circuit
    state = qp.layers.Custom(state, qp.subroutines.gate_builder([qp.subroutines.n_kron(qp.gates.H, buffer), qp.gates.I])) # Apply Hadamard to all the qubits except the output
    state = qp.layers.Measure(state, list(range(buffer))) # Measure all qubits except the output qubit to get the result

    return state[0]


import numpy as np
from collections import defaultdict

class Qubit:
    
    @staticmethod
    def partial_measure(state, meas_qubits):
        n_qubits = int(np.log2(state.shape[0]))
        probs = defaultdict(float)
        post_states = defaultdict(lambda: np.zeros_like(state, dtype=complex))

        for i in range(2 ** n_qubits):
            amp = state[i]
            bits = format(i, f'0{n_qubits}b')
            outcome = ''.join([bits[q] for q in meas_qubits])
            probs[outcome] += np.abs(amp)**2
            post_states[outcome][i] = amp

        states = list(probs.keys())

        measured_state = states[np.random.choice(len(states), p=[i[0] for i in probs.values()])]

        
        norm = np.sqrt(probs[measured_state])
        if norm > 0:
            post_states[measured_state] /= norm

        return measured_state, post_states[measured_state]

    @staticmethod
    def checkValid(qb):
        n = qb.shape[0]
        return (n & (n - 1)) == 0 and sum(np.square(qb)) == 1


import numpy as np

X = np.array(
    [[0, 1],
     [1, 0]]
)

Y = np.array(
    [[0, -1j],
     [1j, 0]]
)

Z = np.array(
    [[1, 0],
     [0, -1]]
)

H = (1/np.sqrt(2)) * np.array(
    [[1, 1],
     [1, -1]]
)

def R(theta):
    return np.array(
    [[np.cos(theta), -np.sin(theta)],
     [np.sin(theta), np.cos(theta)]]
)

S = np.array(
    [[1, 0],
     [0, 1j]]
)

T = np.array(
    [[1, 0],
     [0, np.exp(1j * np.pi / 4)]]
)

def P(a):
    return np.array(
    [[1, 0],
     [0, np.exp(2 * np.pi * 1j / np.power(2,a))]]
)

I = np.eye(2)


import numpy as np
import qupyt as qp

def deutsch(f):

    def deutsch_Uf():
        '''
        Define the Deutsch Unitary
        '''
        nonlocal f
        U = np.zeros((4, 4))
        for x in [0, 1]:
            for y in [0, 1]:
                input_index = 2*x + y
                output_y = y ^ f(x)
                output_index = 2*x + output_y
                U[output_index][input_index] = 1
        return U

    state = qp.layers.Input(2) # Initializes circuit with 2 qubits (all set to ket0)
    state = qp.layers.Custom(state, qp.subroutines.gate_builder([qp.gates.I, qp.gates.X])) # Set the last qubit to ket1
    state = qp.layers.N_Hadamard(state) # Apply Hadamard on all the qubits to get ket+ and ket-
    state = qp.layers.Custom(state, deutsch_Uf()) # Apply the Deutsch unitary to entire circuit
    state = qp.layers.Custom(state, qp.subroutines.gate_builder([qp.gates.H, qp.gates.I])) # Apply Hadamard to only the first qubit
    state = qp.layers.Measure(state, [0]) # Measure the first qubit to get the result

    return 'constant' if int(state[0]) == 0 else 'balanced' # If the measurement is ket0, the function is constant. If ket1, the function is balanced


from setuptools import setup, find_packages

setup(
    name="qupyt",
    version="0.1.3",
    packages=find_packages(),
    install_requires=['numpy>=2.2.0', 'matplotlib>=3.9.0'],
    author="Harikrishna Vardhineedi",
    author_email="harivardhineedi@gmail.com",
    description="Package to emulate and visualize quantum computations with a classical computer.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/SoftLocked/QuPyt",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.10',
)


from qupyt.algorithms.deutsch import deutsch
from qupyt.algorithms.deutsch_josza import deutsch_josza
from qupyt.algorithms.bernstein_vazirani import bernstein_vazirani
from qupyt.algorithms.grover import grover, animate_grover